////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Project   : DevExpress VCL Support
//  * Unit Name : FW_DXDDManager
//  * Purpose   :     TcxPageControl
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2013.
//  * Version   : 1.0.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.ru
//  ****************************************************************************
//

unit FW_DXDDManager;

interface

uses
  Windows,
  Classes,
  Graphics,
  Controls,
  SysUtils,
  Forms,
  cxPC,
  cxControls,
  cxGeometry,
  dxCore,
  dxGDIPlusClasses;

type
  TPageControlData = record
    PageControl: TcxCustomTabControl;
    MouseDown: TMouseEvent;
    MouseMove: TMouseMoveEvent;
    MouseUp: TMouseEvent;
  end;

  TPageRemovedEvent = procedure(Sender: TObject;
    OwnerPageControl: TcxPageControl; Page: TcxTabSheet) of object;

  TDragDropArrow = class(TcxDragImage)
  public
    constructor Create; reintroduce; virtual;
    procedure Init(AOwner: TControl; const AAreaBounds, AClientRect: TRect;
      APlace: TcxArrowPlace); overload;
  end;

  TCustomDragDropManager = class
  private
    FPageControls: array of TPageControlData;
    FPageControlsCount: Integer;
    FMainForm: TForm;
    FSourcePageControl: TcxCustomTabControl;
    FDraggedPage, FBakenPage: TcxTab;
    FDragged, FDropDownRightAtBakenPage: Boolean;
    FDragPointOffset, FStartDragPoint: TPoint;
    FDragImage: TcxDragImage;
    FUpArrow, FDownArrow: TDragDropArrow;
  protected
    procedure MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  protected
    procedure CheckNeedStartDrag(X, Y: Integer);
    function GetPageControlIndex(Control: TObject): Integer;
    function GetPageIndexOnCurrentPageControl(out Control: TcxCustomTabControl): Integer;
    procedure InitDragImages;
    procedure ReleaseDragImages;
    procedure ShowDragImages;
    procedure ShowDragArrow(P: TPoint);
    procedure DropTabTo(ATab: TcxTab; AOldTabControl,
      ANewTabControl: TcxCustomTabControl; ANewIndex: Integer); virtual; abstract;
  public
    constructor Create(MainForm: TForm);
    destructor Destroy; override;
    procedure RegisterPageControl(Value: TcxCustomTabControl); virtual;
  end;

  TDragDropManager = class(TCustomDragDropManager)
  private
    FPageRemoved: TPageRemovedEvent;
  protected
    procedure DropTabTo(ATab: TcxTab; AOldTabControl,
      ANewTabControl: TcxCustomTabControl; ANewIndex: Integer); override;
  public
    procedure RegisterPageControl(Value: TcxPageControl); reintroduce;
    property OnPageRemoved: TPageRemovedEvent read FPageRemoved write FPageRemoved;
  end;

implementation

{$R '.\DragRopManagerRes\ddmanager.res'}

type
  TcxCustomTabControlFriendly = class(TcxCustomTabControl);
  TcxCustomTabControlViewInfoFriendly = class(TcxCustomTabControlViewInfo);
  TcxPCCustomPainterFriendly = class(TcxPCCustomPainter);
  TcxCustomTabControlControllerFriendly = class(TcxCustomTabControlController);

{ TDragDropArrow }

constructor TDragDropArrow.Create;
begin
  inherited Create;
  AlphaBlend := False;
  TransparentColorValue := clWhite;
  TransparentColor := True;
end;

procedure TDragDropArrow.Init(AOwner: TControl;
  const AAreaBounds, AClientRect: TRect; APlace: TcxArrowPlace);
var
  R: TRect;
  I: TdxPNGImage;
begin
  R := GetDragAndDropArrowBounds(AAreaBounds, AClientRect, APlace);
  R := cxRectOffset(R, AOwner.ClientOrigin);
  HandleNeeded;
  BoundsRect := R;
  Canvas.Brush.Color := clWhite;
  Canvas.FillRect(ClientRect);
  I := TdxPNGImage.Create;
  try
    case APlace of
      apTop: I.LoadFromResource(HInstance, 'DD_MANAGER_DOWN_ARROW', 'PNG');
      apBottom: I.LoadFromResource(HInstance, 'DD_MANAGER_UP_ARROW', 'PNG');
    end;
    Canvas.Draw(0, 0, I);
  finally
    I.Free;
  end;
end;

{ TDragDropManager }

procedure TCustomDragDropManager.CheckNeedStartDrag(X, Y: Integer);
begin
  if not FDragged then
  begin
    if X > FStartDragPoint.X then
      X := X - FStartDragPoint.X
    else
      X := FStartDragPoint.X - X;
    if Y > FStartDragPoint.Y then
      Y := Y - FStartDragPoint.Y
    else
      Y := FStartDragPoint.Y - Y;
    FDragged := (X >= 5) or (Y >= 5);
    if FDragged then
    begin
      InitDragImages;
      TcxCustomTabControlFriendly(FSourcePageControl).MouseCapture := True;
    end;
  end;
end;

constructor TCustomDragDropManager.Create(MainForm: TForm);
begin
  FMainForm := MainForm;
end;

destructor TCustomDragDropManager.Destroy;
var
  I: Integer;
  Control: TcxCustomTabControl;
begin
  for I := 0 to FPageControlsCount - 1 do
  begin
    Control := FPageControls[I].PageControl;
    TcxCustomTabControlFriendly(Control).OnMouseDown := FPageControls[I].MouseDown;
    TcxCustomTabControlFriendly(Control).OnMouseMove := FPageControls[I].MouseMove;
    TcxCustomTabControlFriendly(Control).OnMouseUp := FPageControls[I].MouseUp;
  end;
  inherited;
end;

function TCustomDragDropManager.GetPageControlIndex(
  Control: TObject): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to FPageControlsCount - 1 do
    if FPageControls[I].PageControl = Control then
    begin
      Result := I;
      Break;
    end;
end;

function TCustomDragDropManager.GetPageIndexOnCurrentPageControl(
  out Control: TcxCustomTabControl): Integer;

  function TabControlAsPos(Control: TWinControl; const P: TPoint): TcxCustomTabControl;
  var
    LControl: TControl;
  begin
    Result := nil;
    LControl := Control.ControlAtPos(P, True, True);
    if LControl = nil then
      Exit;
    if LControl is TcxCustomTabControl then
    begin
      Result := LControl as TcxCustomTabControl;
      Exit;
    end;
    if LControl is TWinControl then
      Result := TabControlAsPos(LControl as TWinControl, Point(P.X - LControl.Left, P.Y - LControl.Top));
  end;

var
  AControl: TControl;
  P, P1, P2: TPoint;
begin
  Result := -1;
  GetCursorPos(P);
  P1 := FMainForm.ScreenToClient(P);
  AControl := TabControlAsPos(FMainForm, P1);
  if AControl is TcxCustomTabControl then
  begin
    Control := TcxCustomTabControl(AControl);
    P2 := Control.ScreenToClient(P);
    Result := TcxCustomTabControlFriendly(Control).ViewInfo.IndexOfTabAt(P2.X, P2.Y);
  end
  else
    Control := nil;
end;

procedure TCustomDragDropManager.InitDragImages;
var
  R: TRect;
  Index: Integer;
begin
  FDragImage := TcxDragImage.Create;
  Index := FDraggedPage.Index;
  R := FSourcePageControl.ViewInfo.TabsViewInfo[Index].FullRect;
  FDragImage.SetBounds(0, 0, cxRectWidth(R), cxRectHeight(R));
  FDragImage.Canvas.WindowOrg := R.TopLeft;
  FDragImage.Canvas.Canvas.Lock;
  try
    TcxPCCustomPainterFriendly(TcxCustomTabControlViewInfoFriendly(
      FSourcePageControl.ViewInfo).Painter).PaintDragImage(FDragImage.Canvas,
      R, Index);
  finally
    FDragImage.Canvas.Canvas.Unlock;
  end;
end;

procedure TCustomDragDropManager.MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  Index: Integer;
begin
  if Button = mbLeft then
  begin
    FStartDragPoint := Point(X, Y);
    FSourcePageControl := TcxPageControl(Sender);
    Index := FSourcePageControl.ViewInfo.IndexOfTabAt(X, Y);
    if Index >= 0 then
    begin
      FDraggedPage := TcxCustomTabControlFriendly(FSourcePageControl).Tabs[Index];
      FDragPointOffset := cxPointOffset(
        FSourcePageControl.ViewInfo.TabsViewInfo[Index].FullRect.TopLeft,
        Point(X, Y), False);
    end
    else
      FSourcePageControl := nil;
  end;
  Index := GetPageControlIndex(Sender);
  if Index >= 0 then
    if Assigned(FPageControls[Index].MouseDown) then
      FPageControls[Index].MouseDown(Sender, Button, Shift, X, Y);
end;

procedure TCustomDragDropManager.MouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  Index: Integer;
begin
  if FSourcePageControl <> nil then
  begin
    CheckNeedStartDrag(X, Y);
    if FDragged then
      ShowDragImages;
  end;
  Index := GetPageControlIndex(Sender);
  if Index >= 0 then
    if Assigned(FPageControls[Index].MouseMove) then
      FPageControls[Index].MouseMove(Sender, Shift, X, Y);
end;

procedure TCustomDragDropManager.MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  DestinationPageControl: TcxCustomTabControl;
  NewIndex: Integer;
begin
  if Assigned(FDraggedPage) then
  try
    if FDragged then
    begin
      ReleaseDragImages;
      GetPageIndexOnCurrentPageControl(DestinationPageControl);
      if DestinationPageControl <> nil then
      begin
        if FBakenPage = nil then Exit;

        if FSourcePageControl = DestinationPageControl then
        begin
          if FDropDownRightAtBakenPage then
          begin
            if FDraggedPage.Index < FBakenPage.Index then
              NewIndex := FBakenPage.Index
            else
              NewIndex := FBakenPage.Index + 1
          end
          else
            if FDraggedPage.Index < FBakenPage.Index then
              NewIndex := FBakenPage.Index - 1
            else
              NewIndex := FBakenPage.Index;
        end
        else
        begin
          if FDropDownRightAtBakenPage then
            NewIndex := FBakenPage.Index + 1
          else
            NewIndex := FBakenPage.Index;
        end;

        if NewIndex > TcxCustomTabControlFriendly(DestinationPageControl).Tabs.Count then
          NewIndex := TcxCustomTabControlFriendly(DestinationPageControl).Tabs.Count
        else
          if NewIndex < 0 then
            NewIndex := 0;

        DropTabTo(FDraggedPage, FSourcePageControl, DestinationPageControl, NewIndex);
      end;
    end;
  finally
    FSourcePageControl := nil;
    FDraggedPage := nil;
    FDragged := False;
  end;
end;

procedure TCustomDragDropManager.RegisterPageControl(Value: TcxCustomTabControl);
var
  Val: TcxCustomTabControlFriendly;
begin
  if GetPageControlIndex(Value) < 0 then
  begin
    SetLength(FPageControls, FPageControlsCount + 1);
    FPageControls[FPageControlsCount].PageControl := Value;
    Val := TcxCustomTabControlFriendly(Value);
    FPageControls[FPageControlsCount].MouseDown := Val.OnMouseDown;
    FPageControls[FPageControlsCount].MouseMove := Val.OnMouseMove;
    FPageControls[FPageControlsCount].MouseUp := Val.OnMouseUp;
    Val.OnMouseDown := MouseDown;
    Val.OnMouseMove := MouseMove;
    Val.OnMouseUp := MouseUp;
    Inc(FPageControlsCount);
  end;
end;

procedure TCustomDragDropManager.ReleaseDragImages;
begin
  FreeAndNil(FDragImage);
  FreeAndNil(FUpArrow);
  FreeAndNil(FDownArrow);
end;

procedure TCustomDragDropManager.ShowDragArrow(P: TPoint);

  procedure ShowArrow(Hash: Integer;
    Control: TcxCustomTabControl; var Arrow: TDragDropArrow;
    R: TRect; Place: TcxArrowPlace);
  begin
    if Arrow <> nil then
    begin
      if Arrow.Tag = Hash then
        Exit
      else
        FreeAndNil(Arrow);
    end;
    Arrow := TDragDropArrow.Create;
    if Place = apBottom then
      OffsetRect(R, 0, R.Bottom - R.Top + 2)
    else
      OffsetRect(R, 0, -1);
    Arrow.Init(Control, cxNullRect, R, Place);
    Arrow.Tag := Hash;
    Arrow.Visible := True;
  end;

var
  R: TRect;
  Control: TcxCustomTabControl;
  Index, X, Hash: Integer;
  Show: Boolean;
begin
  FBakenPage := nil;
  Index := GetPageIndexOnCurrentPageControl(Control);
  if (Index >= 0) then
  begin

    if Control = FSourcePageControl then
      if Index = FDraggedPage.Index then
        Exit;

    FBakenPage := TcxCustomTabControlFriendly(Control).Tabs.Tabs[Index];

    P := Control.ScreenToClient(P);
    R := Control.ViewInfo.TabsViewInfo[Index].FullRect;
    X := P.X - R.Left;
    FDropDownRightAtBakenPage := ((R.Right - R.Left) shr 1) < X;
    Show := True;
    if FDropDownRightAtBakenPage then
    begin
      OffsetRect(R, R.Right - R.Left, 0);
      if Control = FSourcePageControl then
        Show := (Index + 1 <> FDraggedPage.Index);
    end
    else
    begin
      if Control = FSourcePageControl then
        Show := (Index - 1 <> FDraggedPage.Index);
    end;

    if Show then
    begin
      P := R.TopLeft;
      P := Control.ClientToScreen(P);
      Hash := MakeLong(P.X, P.Y);
      ShowArrow(Hash, Control, FDownArrow, R, apTop);
      ShowArrow(Hash, Control, FUpArrow, R, apBottom);
    end
    else
    begin
      FreeAndNil(FDownArrow);
      FreeAndNil(FUpArrow);
    end;
  end
  else
  begin
    FreeAndNil(FDownArrow);
    FreeAndNil(FUpArrow);
  end;
end;

procedure TCustomDragDropManager.ShowDragImages;
var
  P: TPoint;
begin
  P := GetMouseCursorPos;
  FDragImage.MoveTo(cxPointOffset(P, FDragPointOffset));
  if not FDragImage.Visible then
    FDragImage.Show;
  ShowDragArrow(P);
end;

procedure TDragDropManager.RegisterPageControl(Value: TcxPageControl);
begin
  inherited RegisterPageControl(Value);
end;

{ TDragDropManager }

procedure TDragDropManager.DropTabTo(ATab: TcxTab; AOldTabControl,
  ANewTabControl: TcxCustomTabControl; ANewIndex: Integer);
var
  TabSheet: TcxTabSheet;
begin
  TabSheet := (AOldTabControl as TcxPageControl).Pages[ATab.Index];
  TabSheet.PageControl := ANewTabControl as TcxPageControl;
  TabSheet.PageIndex := ANewIndex;
  if (AOldTabControl <> ANewTabControl) and Assigned(FPageRemoved) then
    FPageRemoved(Self, AOldTabControl as TcxPageControl, TabSheet);
end;

end.
